package org.apache.bcel.generic;

import org.apache.bcel.classfile.LocalVariable;

public class LocalVariableGen implements InstructionTargeter, NamedAndTyped, Cloneable
{
    private int index;
    private String name;
    private Type type;
    private InstructionHandle start;
    private InstructionHandle end;
    private int origIndex;
    private boolean liveToEnd;
    
    public LocalVariableGen(final int index, final String name, final Type type, final InstructionHandle start, final InstructionHandle end) {
        if (index < 0 || index > 65535) {
            throw new ClassGenException("Invalid index: " + index);
        }
        this.name = name;
        this.type = type;
        this.index = index;
        this.setStart(start);
        this.setEnd(end);
        this.origIndex = index;
        this.liveToEnd = (end == null);
    }
    
    public LocalVariableGen(final int index, final String name, final Type type, final InstructionHandle start, final InstructionHandle end, final int origIndex) {
        this(index, name, type, start, end);
        this.origIndex = origIndex;
    }
    
    public Object clone() {
        try {
            return super.clone();
        }
        catch (final CloneNotSupportedException e) {
            throw new Error("Clone Not Supported");
        }
    }
    
    @Override
    public boolean containsTarget(final InstructionHandle ih) {
        return this.start == ih || this.end == ih;
    }
    
    void dispose() {
        this.setStart(null);
        this.setEnd(null);
    }
    
    @Override
    public boolean equals(final Object o) {
        if (!(o instanceof LocalVariableGen)) {
            return false;
        }
        final LocalVariableGen l = (LocalVariableGen)o;
        return l.index == this.index && l.start == this.start && l.end == this.end;
    }
    
    public InstructionHandle getEnd() {
        return this.end;
    }
    
    public int getIndex() {
        return this.index;
    }
    
    public boolean getLiveToEnd() {
        return this.liveToEnd;
    }
    
    public LocalVariable getLocalVariable(final ConstantPoolGen cp) {
        int startPc = 0;
        int length = 0;
        if (this.start != null && this.end != null) {
            startPc = this.start.getPosition();
            length = this.end.getPosition() - startPc;
            if (this.end.getNext() == null && this.liveToEnd) {
                length += this.end.getInstruction().getLength();
            }
        }
        final int nameIndex = cp.addUtf8(this.name);
        final int signatureIndex = cp.addUtf8(this.type.getSignature());
        return new LocalVariable(startPc, length, nameIndex, signatureIndex, this.index, cp.getConstantPool(), this.origIndex);
    }
    
    @Override
    public String getName() {
        return this.name;
    }
    
    public int getOrigIndex() {
        return this.origIndex;
    }
    
    public InstructionHandle getStart() {
        return this.start;
    }
    
    @Override
    public Type getType() {
        return this.type;
    }
    
    @Override
    public int hashCode() {
        return this.name.hashCode() ^ this.type.hashCode();
    }
    
    public void setEnd(final InstructionHandle end) {
        BranchInstruction.notifyTarget(this.end, end, this);
        this.end = end;
    }
    
    public void setIndex(final int index) {
        this.index = index;
    }
    
    public void setLiveToEnd(final boolean liveToEnd) {
        this.liveToEnd = liveToEnd;
    }
    
    @Override
    public void setName(final String name) {
        this.name = name;
    }
    
    public void setStart(final InstructionHandle start) {
        BranchInstruction.notifyTarget(this.start, start, this);
        this.start = start;
    }
    
    @Override
    public void setType(final Type type) {
        this.type = type;
    }
    
    @Override
    public String toString() {
        return "LocalVariableGen(" + this.name + ", " + this.type + ", " + this.start + ", " + this.end + ")";
    }
    
    @Override
    public void updateTarget(final InstructionHandle oldIh, final InstructionHandle newIh) {
        boolean targeted = false;
        if (this.start == oldIh) {
            targeted = true;
            this.setStart(newIh);
        }
        if (this.end == oldIh) {
            targeted = true;
            this.setEnd(newIh);
        }
        if (!targeted) {
            throw new ClassGenException("Not targeting " + oldIh + ", but {" + this.start + ", " + this.end + "}");
        }
    }
}
